@@ -29,6 +29,7 @@
 #include <linux/udp.h>
 #include <linux/icmp.h>
 #include <linux/icmpv6.h>
+#include <linux/rculist.h>
 #include <net/inet_ecn.h>
 #include <net/ip.h>
 #include <net/ipv6.h>
@@ -118,6 +119,67 @@ u64 flow_used_time(unsigned long flow_jiffies)
 	offsetof(struct sw_flow_key, field) +	\
 	FIELD_SIZEOF(struct sw_flow_key, field)
 
+/**
+ * skip_exthdr - skip any IPv6 extension headers
+ * @skb: skbuff to parse
+ * @start: offset of first extension header
+ * @nexthdrp: Initially, points to the type of the extension header at @start.
+ * This function updates it to point to the extension header at the final
+ * offset.
+ * @tos_frag: Points to the @tos_frag member in a &struct sw_flow_key.  This
+ * function sets an appropriate %OVS_FRAG_TYPE_* value.
+ *
+ * This is based on ipv6_skip_exthdr() but adds the updates to *@tos_frag.
+ *
+ * When there is more than one fragment header, this version reports whether
+ * the final fragment header that it examines is a first fragment.
+ *
+ * Returns the final payload offset, or -1 on error.
+ */
+static int skip_exthdr(const struct sk_buff *skb, int start, u8 *nexthdrp,
+		       u8 *tos_frag)
+{
+	u8 nexthdr = *nexthdrp;
+
+	while (ipv6_ext_hdr(nexthdr)) {
+		struct ipv6_opt_hdr _hdr, *hp;
+		int hdrlen;
+
+		if (nexthdr == NEXTHDR_NONE)
+			return -1;
+		hp = skb_header_pointer(skb, start, sizeof(_hdr), &_hdr);
+		if (hp == NULL)
+			return -1;
+		if (nexthdr == NEXTHDR_FRAGMENT) {
+			__be16 _frag_off, *fp;
+			fp = skb_header_pointer(skb,
+						start+offsetof(struct frag_hdr,
+							       frag_off),
+						sizeof(_frag_off),
+						&_frag_off);
+			if (fp == NULL)
+				return -1;
+
+			*tos_frag &= ~OVS_FRAG_TYPE_MASK;
+			if (ntohs(*fp) & ~0x7) {
+				*tos_frag |= OVS_FRAG_TYPE_LATER;
+				break;
+			}
+			*tos_frag |= OVS_FRAG_TYPE_FIRST;
+			hdrlen = 8;
+		} else if (nexthdr == NEXTHDR_AUTH)
+			hdrlen = (hp->hdrlen+2)<<2;
+		else
+			hdrlen = ipv6_optlen(hp);
+
+		nexthdr = hp->nexthdr;
+		start += hdrlen;
+	}
+
+	*nexthdrp = nexthdr;
+	return start;
+}
+
 static int parse_ipv6hdr(struct sk_buff *skb, struct sw_flow_key *key,
 			 int *key_lenp)
 {
@@ -139,11 +201,11 @@ static int parse_ipv6hdr(struct sk_buff *skb, struct sw_flow_key *key,
 	payload_ofs = (u8 *)(nh + 1) - skb->data;
 
 	key->ip.proto = NEXTHDR_NONE;
-	key->ip.tos = ipv6_get_dsfield(nh) & ~INET_ECN_MASK;
+	key->ip.tos_frag = ipv6_get_dsfield(nh) & ~INET_ECN_MASK;
 	ipv6_addr_copy(&key->ipv6.addr.src, &nh->saddr);
 	ipv6_addr_copy(&key->ipv6.addr.dst, &nh->daddr);
 
-	payload_ofs = ipv6_skip_exthdr(skb, payload_ofs, &nexthdr);
+	payload_ofs = skip_exthdr(skb, payload_ofs, &nexthdr, &key->ip.tos_frag);
 	if (unlikely(payload_ofs < 0))
 		return -EINVAL;
 
@@ -172,12 +234,12 @@ void flow_used(struct sw_flow *flow, struct sk_buff *skb)
 		tcp_flags = *(tcp + TCP_FLAGS_OFFSET) & TCP_FLAG_MASK;
 	}
 
-	spin_lock_bh(&flow->lock);
+	spin_lock(&flow->lock);
 	flow->used = jiffies;
 	flow->packet_count++;
 	flow->byte_count += skb->len;
 	flow->tcp_flags |= tcp_flags;
-	spin_unlock_bh(&flow->lock);
+	spin_unlock(&flow->lock);
 }
 
 struct sw_flow_actions *flow_actions_alloc(const struct nlattr *actions)
@@ -216,14 +278,153 @@ struct sw_flow *flow_alloc(void)
 	return flow;
 }
 
-void flow_free_tbl(struct tbl_node *node)
+static struct hlist_head __rcu *find_bucket(struct flow_table * table, u32 hash)
 {
-	struct sw_flow *flow = flow_cast(node);
+	return flex_array_get(table->buckets,
+				(hash & (table->n_buckets - 1)));
+}
 
+static struct flex_array  __rcu *alloc_buckets(unsigned int n_buckets)
+{
+	struct flex_array  __rcu * buckets;
+	int i, err;
+
+	buckets = flex_array_alloc(sizeof(struct hlist_head *),
+				   n_buckets, GFP_KERNEL);
+	if (!buckets)
+		return NULL;
+
+	err = flex_array_prealloc(buckets, 0, n_buckets, GFP_KERNEL);
+	if (err) {
+		flex_array_free(buckets);
+		return NULL;
+	}
+
+	for (i = 0; i < n_buckets; i++)
+		INIT_HLIST_HEAD((struct hlist_head *)
+					flex_array_get(buckets, i));
+
+	return buckets;
+}
+
+static void free_buckets(struct flex_array * buckets)
+{
+	flex_array_free(buckets);
+}
+
+struct flow_table *flow_tbl_alloc(int new_size)
+{
+	struct flow_table *table = kmalloc(sizeof(*table), GFP_KERNEL);
+
+	if (!table)
+		return NULL;
+
+	table->buckets = alloc_buckets(new_size);
+
+	if (!table->buckets) {
+		kfree(table);
+		return NULL;
+	}
+	table->n_buckets = new_size;
+	table->count = 0;
+
+	return table;
+}
+
+static void flow_free(struct sw_flow *flow)
+{
 	flow->dead = true;
 	flow_put(flow);
 }
 
+void flow_tbl_destroy(struct flow_table *table)
+{
+	int i;
+
+	if (!table)
+		return;
+
+	for (i = 0; i < table->n_buckets; i++) {
+		struct sw_flow *flow;
+		struct hlist_head *head = flex_array_get(table->buckets, i);
+		struct hlist_node *node, *n;
+
+		hlist_for_each_entry_safe(flow, node, n, head, hash_node) {
+			hlist_del_init_rcu(&flow->hash_node);
+			flow_free(flow);
+		}
+	}
+
+	free_buckets(table->buckets);
+	kfree(table);
+}
+
+static void flow_tbl_destroy_rcu_cb(struct rcu_head *rcu)
+{
+	struct flow_table *table = container_of(rcu, struct flow_table, rcu);
+
+	flow_tbl_destroy(table);
+}
+
+void flow_tbl_deferred_destroy(struct flow_table *table)
+{
+        if (!table)
+                return;
+
+        call_rcu(&table->rcu, flow_tbl_destroy_rcu_cb);
+}
+
+struct sw_flow *flow_tbl_next(struct flow_table *table, u32 *bucket, u32 *last)
+{
+	struct sw_flow *flow;
+	struct hlist_head *head;
+	struct hlist_node *n;
+	int i;
+
+	while (*bucket < table->n_buckets) {
+		i = 0;
+		head = flex_array_get(table->buckets, *bucket);
+		hlist_for_each_entry_rcu(flow, n, head, hash_node) {
+			if (i < *last) {
+				i++;
+				continue;
+			}
+			*last = i + 1;
+			return flow;
+		}
+		(*bucket)++;
+		*last = 0;
+	}
+
+	return NULL;
+}
+
+struct flow_table *flow_tbl_expand(struct flow_table *table)
+{
+	struct flow_table *new_table;
+	int n_buckets = table->n_buckets * 2;
+	int i;
+
+	new_table = flow_tbl_alloc(n_buckets);
+	if (!new_table)
+		return ERR_PTR(-ENOMEM);
+
+	for (i = 0; i < table->n_buckets; i++) {
+		struct sw_flow *flow;
+		struct hlist_head *head;
+		struct hlist_node *n, *pos;
+
+		head = flex_array_get(table->buckets, i);
+
+		hlist_for_each_entry_safe(flow, n, pos, head, hash_node) {
+			hlist_del_init_rcu(&flow->hash_node);
+			flow_tbl_insert(new_table, flow);
+		}
+	}
+
+	return new_table;
+}
+
 /* RCU callback used by flow_deferred_free. */
 static void rcu_free_flow_callback(struct rcu_head *rcu)
 {
@@ -412,8 +613,6 @@ out:
  * @in_port: port number on which @skb was received.
  * @key: output flow key
  * @key_lenp: length of output flow key
- * @is_frag: set to 1 if @skb contains an IPv4 fragment, or to 0 if @skb does
- * not contain an IPv4 packet or if it is not a fragment.
  *
  * The caller must ensure that skb->len >= ETH_HLEN.
  *
@@ -432,7 +631,7 @@ out:
  *      For other key->dl_type values it is left untouched.
  */
 int flow_extract(struct sk_buff *skb, u16 in_port, struct sw_flow_key *key,
-		 int *key_lenp, bool *is_frag)
+		 int *key_lenp)
 {
 	int error = 0;
 	int key_len = SW_FLOW_KEY_OFFSET(eth);
@@ -441,7 +640,6 @@ int flow_extract(struct sk_buff *skb, u16 in_port, struct sw_flow_key *key,
 	memset(key, 0, sizeof(*key));
 	key->eth.tun_id = OVS_CB(skb)->tun_id;
 	key->eth.in_port = in_port;
-	*is_frag = false;
 
 	skb_reset_mac_header(skb);
 
@@ -470,6 +668,7 @@ int flow_extract(struct sk_buff *skb, u16 in_port, struct sw_flow_key *key,
 	/* Network layer. */
 	if (key->eth.type == htons(ETH_P_IP)) {
 		struct iphdr *nh;
+		__be16 offset;
 
 		key_len = SW_FLOW_KEY_OFFSET(ipv4.addr);
 
@@ -485,31 +684,37 @@ int flow_extract(struct sk_buff *skb, u16 in_port, struct sw_flow_key *key,
 		nh = ip_hdr(skb);
 		key->ipv4.addr.src = nh->saddr;
 		key->ipv4.addr.dst = nh->daddr;
-		key->ip.tos = nh->tos & ~INET_ECN_MASK;
+
 		key->ip.proto = nh->protocol;
+		key->ip.tos_frag = nh->tos & ~INET_ECN_MASK;
+
+		offset = nh->frag_off & htons(IP_OFFSET);
+		if (offset) {
+			key->ip.tos_frag |= OVS_FRAG_TYPE_LATER;
+			goto out;
+		}
+		if (nh->frag_off & htons(IP_MF) ||
+			 skb_shinfo(skb)->gso_type & SKB_GSO_UDP)
+			key->ip.tos_frag |= OVS_FRAG_TYPE_FIRST;
 
 		/* Transport layer. */
-		if ((nh->frag_off & htons(IP_MF | IP_OFFSET)) ||
-		    (skb_shinfo(skb)->gso_type & SKB_GSO_UDP))
-			*is_frag = true;
-
 		if (key->ip.proto == IPPROTO_TCP) {
 			key_len = SW_FLOW_KEY_OFFSET(ipv4.tp);
-			if (!*is_frag && tcphdr_ok(skb)) {
+			if (tcphdr_ok(skb)) {
 				struct tcphdr *tcp = tcp_hdr(skb);
 				key->ipv4.tp.src = tcp->source;
 				key->ipv4.tp.dst = tcp->dest;
 			}
 		} else if (key->ip.proto == IPPROTO_UDP) {
 			key_len = SW_FLOW_KEY_OFFSET(ipv4.tp);
-			if (!*is_frag && udphdr_ok(skb)) {
+			if (udphdr_ok(skb)) {
 				struct udphdr *udp = udp_hdr(skb);
 				key->ipv4.tp.src = udp->source;
 				key->ipv4.tp.dst = udp->dest;
 			}
 		} else if (key->ip.proto == IPPROTO_ICMP) {
 			key_len = SW_FLOW_KEY_OFFSET(ipv4.tp);
-			if (!*is_frag && icmphdr_ok(skb)) {
+			if (icmphdr_ok(skb)) {
 				struct icmphdr *icmp = icmp_hdr(skb);
 				/* The ICMP type and code fields use the 16-bit
 				 * transport port fields, so we need to store them
@@ -554,6 +759,11 @@ int flow_extract(struct sk_buff *skb, u16 in_port, struct sw_flow_key *key,
 			goto out;
 		}
 
+		if ((key->ip.tos_frag & OVS_FRAG_TYPE_MASK) == OVS_FRAG_TYPE_LATER)
+			goto out;
+		if (skb_shinfo(skb)->gso_type & SKB_GSO_UDP)
+			key->ip.tos_frag |= OVS_FRAG_TYPE_FIRST;
+
 		/* Transport layer. */
 		if (key->ip.proto == NEXTHDR_TCP) {
 			key_len = SW_FLOW_KEY_OFFSET(ipv6.tp);
@@ -589,36 +799,76 @@ u32 flow_hash(const struct sw_flow_key *key, int key_len)
 	return jhash2((u32*)key, DIV_ROUND_UP(key_len, sizeof(u32)), hash_seed);
 }
 
-int flow_cmp(const struct tbl_node *node, void *key2_, int len)
+struct sw_flow * flow_tbl_lookup(struct flow_table *table,
+				struct sw_flow_key *key, int key_len)
 {
-	const struct sw_flow_key *key1 = &flow_cast(node)->key;
-	const struct sw_flow_key *key2 = key2_;
+	struct sw_flow *flow;
+	struct hlist_node *n;
+	struct hlist_head *head;
+	u32 hash;
 
-	return !memcmp(key1, key2, len);
+	hash = flow_hash(key, key_len);
+
+	head = find_bucket(table, hash);
+	hlist_for_each_entry_rcu(flow, n, head, hash_node) {
+
+		if (flow->hash == hash &&
+		    !memcmp(&flow->key, key, key_len)) {
+			return flow;
+		}
+	}
+	return NULL;
 }
 
-/* The size of the argument for each %ODP_KEY_ATTR_* Netlink attribute.  */
-static const u32 key_lens[ODP_KEY_ATTR_MAX + 1] = {
-	[ODP_KEY_ATTR_TUN_ID] = 8,
-	[ODP_KEY_ATTR_IN_PORT] = 4,
-	[ODP_KEY_ATTR_ETHERNET] = sizeof(struct odp_key_ethernet),
-	[ODP_KEY_ATTR_8021Q] = sizeof(struct odp_key_8021q),
-	[ODP_KEY_ATTR_ETHERTYPE] = 2,
-	[ODP_KEY_ATTR_IPV4] = sizeof(struct odp_key_ipv4),
-	[ODP_KEY_ATTR_IPV6] = sizeof(struct odp_key_ipv6),
-	[ODP_KEY_ATTR_TCP] = sizeof(struct odp_key_tcp),
-	[ODP_KEY_ATTR_UDP] = sizeof(struct odp_key_udp),
-	[ODP_KEY_ATTR_ICMP] = sizeof(struct odp_key_icmp),
-	[ODP_KEY_ATTR_ICMPV6] = sizeof(struct odp_key_icmpv6),
-	[ODP_KEY_ATTR_ARP] = sizeof(struct odp_key_arp),
-	[ODP_KEY_ATTR_ND] = sizeof(struct odp_key_nd),
+void flow_tbl_insert(struct flow_table *table, struct sw_flow *flow)
+{
+	struct hlist_head *head;
+
+	head = find_bucket(table, flow->hash);
+	hlist_add_head_rcu(&flow->hash_node, head);
+	table->count++;
+}
+
+void flow_tbl_remove(struct flow_table *table, struct sw_flow *flow)
+{
+	if (!hlist_unhashed(&flow->hash_node)) {
+		hlist_del_init_rcu(&flow->hash_node);
+		table->count--;
+		BUG_ON(table->count < 0);
+	}
+}
+
+static int parse_tos_frag(struct sw_flow_key *swkey, u8 tos, u8 frag)
+{
+	if (tos & INET_ECN_MASK || frag > OVS_FRAG_TYPE_MAX)
+		return -EINVAL;
+
+	swkey->ip.tos_frag = tos | frag;
+	return 0;
+}
+
+/* The size of the argument for each %OVS_KEY_ATTR_* Netlink attribute.  */
+const u32 ovs_key_lens[OVS_KEY_ATTR_MAX + 1] = {
+	[OVS_KEY_ATTR_TUN_ID] = 8,
+	[OVS_KEY_ATTR_IN_PORT] = 4,
+	[OVS_KEY_ATTR_ETHERNET] = sizeof(struct ovs_key_ethernet),
+	[OVS_KEY_ATTR_8021Q] = sizeof(struct ovs_key_8021q),
+	[OVS_KEY_ATTR_ETHERTYPE] = 2,
+	[OVS_KEY_ATTR_IPV4] = sizeof(struct ovs_key_ipv4),
+	[OVS_KEY_ATTR_IPV6] = sizeof(struct ovs_key_ipv6),
+	[OVS_KEY_ATTR_TCP] = sizeof(struct ovs_key_tcp),
+	[OVS_KEY_ATTR_UDP] = sizeof(struct ovs_key_udp),
+	[OVS_KEY_ATTR_ICMP] = sizeof(struct ovs_key_icmp),
+	[OVS_KEY_ATTR_ICMPV6] = sizeof(struct ovs_key_icmpv6),
+	[OVS_KEY_ATTR_ARP] = sizeof(struct ovs_key_arp),
+	[OVS_KEY_ATTR_ND] = sizeof(struct ovs_key_nd),
 };
 
 /**
  * flow_from_nlattrs - parses Netlink attributes into a flow key.
  * @swkey: receives the extracted flow key.
  * @key_lenp: number of bytes used in @swkey.
- * @attr: Netlink attribute holding nested %ODP_KEY_ATTR_* Netlink attribute
+ * @attr: Netlink attribute holding nested %OVS_KEY_ATTR_* Netlink attribute
  * sequence.
  *
  * This state machine accepts the following forms, with [] for optional
@@ -626,11 +876,15 @@ static const u32 key_lens[ODP_KEY_ATTR_MAX + 1] = {
  *
  * [tun_id] [in_port] ethernet [8021q] [ethertype \
  *              [IPv4 [TCP|UDP|ICMP] | IPv6 [TCP|UDP|ICMPv6 [ND]] | ARP]]
+ *
+ * except that IPv4 or IPv6 terminates the sequence if its @ipv4_frag or
+ * @ipv6_frag member, respectively, equals %OVS_FRAG_TYPE_LATER.
  */
 int flow_from_nlattrs(struct sw_flow_key *swkey, int *key_lenp,
 		      const struct nlattr *attr)
 {
 	int error = 0;
+	enum ovs_frag_type frag_type;
 	const struct nlattr *nla;
 	u16 prev_type;
 	int rem;
@@ -641,46 +895,46 @@ int flow_from_nlattrs(struct sw_flow_key *swkey, int *key_lenp,
 	swkey->eth.type = htons(ETH_P_802_2);
 	key_len = SW_FLOW_KEY_OFFSET(eth);
 
-	prev_type = ODP_KEY_ATTR_UNSPEC;
+	prev_type = OVS_KEY_ATTR_UNSPEC;
 	nla_for_each_nested(nla, attr, rem) {
-		const struct odp_key_ethernet *eth_key;
-		const struct odp_key_8021q *q_key;
-		const struct odp_key_ipv4 *ipv4_key;
-		const struct odp_key_ipv6 *ipv6_key;
-		const struct odp_key_tcp *tcp_key;
-		const struct odp_key_udp *udp_key;
-		const struct odp_key_icmp *icmp_key;
-		const struct odp_key_icmpv6 *icmpv6_key;
-		const struct odp_key_arp *arp_key;
-		const struct odp_key_nd *nd_key;
+		const struct ovs_key_ethernet *eth_key;
+		const struct ovs_key_8021q *q_key;
+		const struct ovs_key_ipv4 *ipv4_key;
+		const struct ovs_key_ipv6 *ipv6_key;
+		const struct ovs_key_tcp *tcp_key;
+		const struct ovs_key_udp *udp_key;
+		const struct ovs_key_icmp *icmp_key;
+		const struct ovs_key_icmpv6 *icmpv6_key;
+		const struct ovs_key_arp *arp_key;
+		const struct ovs_key_nd *nd_key;
 
                 int type = nla_type(nla);
 
-                if (type > ODP_KEY_ATTR_MAX || nla_len(nla) != key_lens[type])
-                        goto invalid;
+		if (type > OVS_KEY_ATTR_MAX || nla_len(nla) != ovs_key_lens[type])
+			goto invalid;
 
 #define TRANSITION(PREV_TYPE, TYPE) (((PREV_TYPE) << 16) | (TYPE))
 		switch (TRANSITION(prev_type, type)) {
-		case TRANSITION(ODP_KEY_ATTR_UNSPEC, ODP_KEY_ATTR_TUN_ID):
+		case TRANSITION(OVS_KEY_ATTR_UNSPEC, OVS_KEY_ATTR_TUN_ID):
 			swkey->eth.tun_id = nla_get_be64(nla);
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_UNSPEC, ODP_KEY_ATTR_IN_PORT):
-		case TRANSITION(ODP_KEY_ATTR_TUN_ID, ODP_KEY_ATTR_IN_PORT):
+		case TRANSITION(OVS_KEY_ATTR_UNSPEC, OVS_KEY_ATTR_IN_PORT):
+		case TRANSITION(OVS_KEY_ATTR_TUN_ID, OVS_KEY_ATTR_IN_PORT):
 			if (nla_get_u32(nla) >= DP_MAX_PORTS)
 				goto invalid;
 			swkey->eth.in_port = nla_get_u32(nla);
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_UNSPEC, ODP_KEY_ATTR_ETHERNET):
-		case TRANSITION(ODP_KEY_ATTR_TUN_ID, ODP_KEY_ATTR_ETHERNET):
-		case TRANSITION(ODP_KEY_ATTR_IN_PORT, ODP_KEY_ATTR_ETHERNET):
+		case TRANSITION(OVS_KEY_ATTR_UNSPEC, OVS_KEY_ATTR_ETHERNET):
+		case TRANSITION(OVS_KEY_ATTR_TUN_ID, OVS_KEY_ATTR_ETHERNET):
+		case TRANSITION(OVS_KEY_ATTR_IN_PORT, OVS_KEY_ATTR_ETHERNET):
 			eth_key = nla_data(nla);
 			memcpy(swkey->eth.src, eth_key->eth_src, ETH_ALEN);
 			memcpy(swkey->eth.dst, eth_key->eth_dst, ETH_ALEN);
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_ETHERNET, ODP_KEY_ATTR_8021Q):
+		case TRANSITION(OVS_KEY_ATTR_ETHERNET, OVS_KEY_ATTR_8021Q):
 			q_key = nla_data(nla);
 			/* Only standard 0x8100 VLANs currently supported. */
 			if (q_key->q_tpid != htons(ETH_P_8021Q))
@@ -690,42 +944,42 @@ int flow_from_nlattrs(struct sw_flow_key *swkey, int *key_lenp,
 			swkey->eth.tci = q_key->q_tci | htons(VLAN_TAG_PRESENT);
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_8021Q, ODP_KEY_ATTR_ETHERTYPE):
-		case TRANSITION(ODP_KEY_ATTR_ETHERNET, ODP_KEY_ATTR_ETHERTYPE):
+		case TRANSITION(OVS_KEY_ATTR_8021Q, OVS_KEY_ATTR_ETHERTYPE):
+		case TRANSITION(OVS_KEY_ATTR_ETHERNET, OVS_KEY_ATTR_ETHERTYPE):
 			swkey->eth.type = nla_get_be16(nla);
 			if (ntohs(swkey->eth.type) < 1536)
 				goto invalid;
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_ETHERTYPE, ODP_KEY_ATTR_IPV4):
+		case TRANSITION(OVS_KEY_ATTR_ETHERTYPE, OVS_KEY_ATTR_IPV4):
 			key_len = SW_FLOW_KEY_OFFSET(ipv4.addr);
 			if (swkey->eth.type != htons(ETH_P_IP))
 				goto invalid;
 			ipv4_key = nla_data(nla);
 			swkey->ip.proto = ipv4_key->ipv4_proto;
-			swkey->ip.tos = ipv4_key->ipv4_tos;
+			if (parse_tos_frag(swkey, ipv4_key->ipv4_tos,
+					   ipv4_key->ipv4_frag))
+				goto invalid;
 			swkey->ipv4.addr.src = ipv4_key->ipv4_src;
 			swkey->ipv4.addr.dst = ipv4_key->ipv4_dst;
-			if (swkey->ip.tos & INET_ECN_MASK)
-				goto invalid;
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_ETHERTYPE, ODP_KEY_ATTR_IPV6):
+		case TRANSITION(OVS_KEY_ATTR_ETHERTYPE, OVS_KEY_ATTR_IPV6):
 			key_len = SW_FLOW_KEY_OFFSET(ipv6.addr);
 			if (swkey->eth.type != htons(ETH_P_IPV6))
 				goto invalid;
 			ipv6_key = nla_data(nla);
 			swkey->ip.proto = ipv6_key->ipv6_proto;
-			swkey->ip.tos = ipv6_key->ipv6_tos;
+			if (parse_tos_frag(swkey, ipv6_key->ipv6_tos,
+					   ipv6_key->ipv6_frag))
+				goto invalid;
 			memcpy(&swkey->ipv6.addr.src, ipv6_key->ipv6_src,
 					sizeof(swkey->ipv6.addr.src));
 			memcpy(&swkey->ipv6.addr.dst, ipv6_key->ipv6_dst,
 					sizeof(swkey->ipv6.addr.dst));
-			if (swkey->ip.tos & INET_ECN_MASK)
-				goto invalid;
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_IPV4, ODP_KEY_ATTR_TCP):
+		case TRANSITION(OVS_KEY_ATTR_IPV4, OVS_KEY_ATTR_TCP):
 			key_len = SW_FLOW_KEY_OFFSET(ipv4.tp);
 			if (swkey->ip.proto != IPPROTO_TCP)
 				goto invalid;
@@ -734,7 +988,7 @@ int flow_from_nlattrs(struct sw_flow_key *swkey, int *key_lenp,
 			swkey->ipv4.tp.dst = tcp_key->tcp_dst;
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_IPV6, ODP_KEY_ATTR_TCP):
+		case TRANSITION(OVS_KEY_ATTR_IPV6, OVS_KEY_ATTR_TCP):
 			key_len = SW_FLOW_KEY_OFFSET(ipv6.tp);
 			if (swkey->ip.proto != IPPROTO_TCP)
 				goto invalid;
@@ -743,7 +997,7 @@ int flow_from_nlattrs(struct sw_flow_key *swkey, int *key_lenp,
 			swkey->ipv6.tp.dst = tcp_key->tcp_dst;
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_IPV4, ODP_KEY_ATTR_UDP):
+		case TRANSITION(OVS_KEY_ATTR_IPV4, OVS_KEY_ATTR_UDP):
 			key_len = SW_FLOW_KEY_OFFSET(ipv4.tp);
 			if (swkey->ip.proto != IPPROTO_UDP)
 				goto invalid;
@@ -752,7 +1006,7 @@ int flow_from_nlattrs(struct sw_flow_key *swkey, int *key_lenp,
 			swkey->ipv4.tp.dst = udp_key->udp_dst;
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_IPV6, ODP_KEY_ATTR_UDP):
+		case TRANSITION(OVS_KEY_ATTR_IPV6, OVS_KEY_ATTR_UDP):
 			key_len = SW_FLOW_KEY_OFFSET(ipv6.tp);
 			if (swkey->ip.proto != IPPROTO_UDP)
 				goto invalid;
@@ -761,7 +1015,7 @@ int flow_from_nlattrs(struct sw_flow_key *swkey, int *key_lenp,
 			swkey->ipv6.tp.dst = udp_key->udp_dst;
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_IPV4, ODP_KEY_ATTR_ICMP):
+		case TRANSITION(OVS_KEY_ATTR_IPV4, OVS_KEY_ATTR_ICMP):
 			key_len = SW_FLOW_KEY_OFFSET(ipv4.tp);
 			if (swkey->ip.proto != IPPROTO_ICMP)
 				goto invalid;
@@ -770,7 +1024,7 @@ int flow_from_nlattrs(struct sw_flow_key *swkey, int *key_lenp,
 			swkey->ipv4.tp.dst = htons(icmp_key->icmp_code);
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_IPV6, ODP_KEY_ATTR_ICMPV6):
+		case TRANSITION(OVS_KEY_ATTR_IPV6, OVS_KEY_ATTR_ICMPV6):
 			key_len = SW_FLOW_KEY_OFFSET(ipv6.tp);
 			if (swkey->ip.proto != IPPROTO_ICMPV6)
 				goto invalid;
@@ -779,7 +1033,7 @@ int flow_from_nlattrs(struct sw_flow_key *swkey, int *key_lenp,
 			swkey->ipv6.tp.dst = htons(icmpv6_key->icmpv6_code);
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_ETHERTYPE, ODP_KEY_ATTR_ARP):
+		case TRANSITION(OVS_KEY_ATTR_ETHERTYPE, OVS_KEY_ATTR_ARP):
 			key_len = SW_FLOW_KEY_OFFSET(ipv4.arp);
 			if (swkey->eth.type != htons(ETH_P_ARP))
 				goto invalid;
@@ -793,7 +1047,7 @@ int flow_from_nlattrs(struct sw_flow_key *swkey, int *key_lenp,
 			memcpy(swkey->ipv4.arp.tha, arp_key->arp_tha, ETH_ALEN);
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_ICMPV6, ODP_KEY_ATTR_ND):
+		case TRANSITION(OVS_KEY_ATTR_ICMPV6, OVS_KEY_ATTR_ND):
 			key_len = SW_FLOW_KEY_OFFSET(ipv6.nd);
 			if (swkey->ipv6.tp.src != htons(NDISC_NEIGHBOUR_SOLICITATION)
 			    && swkey->ipv6.tp.src != htons(NDISC_NEIGHBOUR_ADVERTISEMENT))
@@ -814,49 +1068,59 @@ int flow_from_nlattrs(struct sw_flow_key *swkey, int *key_lenp,
 	if (rem)
 		goto invalid;
 
+	frag_type = swkey->ip.tos_frag & OVS_FRAG_TYPE_MASK;
 	switch (prev_type) {
-	case ODP_KEY_ATTR_UNSPEC:
+	case OVS_KEY_ATTR_UNSPEC:
 		goto invalid;
 
-	case ODP_KEY_ATTR_TUN_ID:
-	case ODP_KEY_ATTR_IN_PORT:
+	case OVS_KEY_ATTR_TUN_ID:
+	case OVS_KEY_ATTR_IN_PORT:
 		goto invalid;
 
-	case ODP_KEY_ATTR_ETHERNET:
-	case ODP_KEY_ATTR_8021Q:
+	case OVS_KEY_ATTR_ETHERNET:
+	case OVS_KEY_ATTR_8021Q:
 		goto ok;
 
-	case ODP_KEY_ATTR_ETHERTYPE:
+	case OVS_KEY_ATTR_ETHERTYPE:
 		if (swkey->eth.type == htons(ETH_P_IP) ||
 		    swkey->eth.type == htons(ETH_P_ARP))
 			goto invalid;
 		goto ok;
 
-	case ODP_KEY_ATTR_IPV4:
+	case OVS_KEY_ATTR_IPV4:
+		if (frag_type == OVS_FRAG_TYPE_LATER)
+			goto ok;
 		if (swkey->ip.proto == IPPROTO_TCP ||
 		    swkey->ip.proto == IPPROTO_UDP ||
 		    swkey->ip.proto == IPPROTO_ICMP)
 			goto invalid;
 		goto ok;
 
-	case ODP_KEY_ATTR_IPV6:
+	case OVS_KEY_ATTR_IPV6:
+		if (frag_type == OVS_FRAG_TYPE_LATER)
+			goto ok;
 		if (swkey->ip.proto == IPPROTO_TCP ||
 		    swkey->ip.proto == IPPROTO_UDP ||
 		    swkey->ip.proto == IPPROTO_ICMPV6)
 			goto invalid;
 		goto ok;
 
-	case ODP_KEY_ATTR_ICMPV6:
+	case OVS_KEY_ATTR_ICMPV6:
 		if (swkey->ipv6.tp.src == htons(NDISC_NEIGHBOUR_SOLICITATION) ||
-		    swkey->ipv6.tp.src == htons(NDISC_NEIGHBOUR_ADVERTISEMENT))
+		    swkey->ipv6.tp.src == htons(NDISC_NEIGHBOUR_ADVERTISEMENT) ||
+		    frag_type == OVS_FRAG_TYPE_LATER)
 			goto invalid;
 		goto ok;
 
-	case ODP_KEY_ATTR_TCP:
-	case ODP_KEY_ATTR_UDP:
-	case ODP_KEY_ATTR_ICMP:
-	case ODP_KEY_ATTR_ARP:
-	case ODP_KEY_ATTR_ND:
+	case OVS_KEY_ATTR_TCP:
+	case OVS_KEY_ATTR_UDP:
+	case OVS_KEY_ATTR_ICMP:
+	case OVS_KEY_ATTR_ND:
+		if (frag_type == OVS_FRAG_TYPE_LATER)
+			goto invalid;
+		goto ok;
+
+	case OVS_KEY_ATTR_ARP:
 		goto ok;
 
 	default:
@@ -876,7 +1140,7 @@ ok:
  * flow_metadata_from_nlattrs - parses Netlink attributes into a flow key.
  * @in_port: receives the extracted input port.
  * @tun_id: receives the extracted tunnel ID.
- * @key: Netlink attribute holding nested %ODP_KEY_ATTR_* Netlink attribute
+ * @key: Netlink attribute holding nested %OVS_KEY_ATTR_* Netlink attribute
  * sequence.
  *
  * This parses a series of Netlink attributes that form a flow key, which must
@@ -894,20 +1158,20 @@ int flow_metadata_from_nlattrs(u16 *in_port, __be64 *tun_id,
 	*in_port = USHRT_MAX;
 	*tun_id = 0;
 
-	prev_type = ODP_KEY_ATTR_UNSPEC;
+	prev_type = OVS_KEY_ATTR_UNSPEC;
 	nla_for_each_nested(nla, attr, rem) {
                 int type = nla_type(nla);
 
-                if (type > ODP_KEY_ATTR_MAX || nla_len(nla) != key_lens[type])
-                        return -EINVAL;
+		if (type > OVS_KEY_ATTR_MAX || nla_len(nla) != ovs_key_lens[type])
+			return -EINVAL;
 
 		switch (TRANSITION(prev_type, type)) {
-		case TRANSITION(ODP_KEY_ATTR_UNSPEC, ODP_KEY_ATTR_TUN_ID):
+		case TRANSITION(OVS_KEY_ATTR_UNSPEC, OVS_KEY_ATTR_TUN_ID):
 			*tun_id = nla_get_be64(nla);
 			break;
 
-		case TRANSITION(ODP_KEY_ATTR_UNSPEC, ODP_KEY_ATTR_IN_PORT):
-		case TRANSITION(ODP_KEY_ATTR_TUN_ID, ODP_KEY_ATTR_IN_PORT):
+		case TRANSITION(OVS_KEY_ATTR_UNSPEC, OVS_KEY_ATTR_IN_PORT):
+		case TRANSITION(OVS_KEY_ATTR_TUN_ID, OVS_KEY_ATTR_IN_PORT):
 			if (nla_get_u32(nla) >= DP_MAX_PORTS)
 				return -EINVAL;
 			*in_port = nla_get_u32(nla);
@@ -921,27 +1185,26 @@ int flow_metadata_from_nlattrs(u16 *in_port, __be64 *tun_id,
 	}
 	if (rem)
 		return -EINVAL;
-
 	return 0;
 }
 
 int flow_to_nlattrs(const struct sw_flow_key *swkey, struct sk_buff *skb)
 {
-	struct odp_key_ethernet *eth_key;
+	struct ovs_key_ethernet *eth_key;
 	struct nlattr *nla;
 
 	/* This is an imperfect sanity-check that FLOW_BUFSIZE doesn't need
-	 * to be updated, but will at least raise awareness when new ODP key
-	 * types are added. */
-	BUILD_BUG_ON(__ODP_KEY_ATTR_MAX != 14);
+	 * to be updated, but will at least raise awareness when new
+	 * datapath key types are added. */
+	BUILD_BUG_ON(__OVS_KEY_ATTR_MAX != 14);
 
 	if (swkey->eth.tun_id != cpu_to_be64(0))
-		NLA_PUT_BE64(skb, ODP_KEY_ATTR_TUN_ID, swkey->eth.tun_id);
+		NLA_PUT_BE64(skb, OVS_KEY_ATTR_TUN_ID, swkey->eth.tun_id);
 
 	if (swkey->eth.in_port != USHRT_MAX)
-		NLA_PUT_U32(skb, ODP_KEY_ATTR_IN_PORT, swkey->eth.in_port);
+		NLA_PUT_U32(skb, OVS_KEY_ATTR_IN_PORT, swkey->eth.in_port);
 
-	nla = nla_reserve(skb, ODP_KEY_ATTR_ETHERNET, sizeof(*eth_key));
+	nla = nla_reserve(skb, OVS_KEY_ATTR_ETHERNET, sizeof(*eth_key));
 	if (!nla)
 		goto nla_put_failure;
 	eth_key = nla_data(nla);
@@ -949,52 +1212,54 @@ int flow_to_nlattrs(const struct sw_flow_key *swkey, struct sk_buff *skb)
 	memcpy(eth_key->eth_dst, swkey->eth.dst, ETH_ALEN);
 
 	if (swkey->eth.tci != htons(0)) {
-		struct odp_key_8021q q_key;
+		struct ovs_key_8021q q_key;
 
 		q_key.q_tpid = htons(ETH_P_8021Q);
 		q_key.q_tci = swkey->eth.tci & ~htons(VLAN_TAG_PRESENT);
-		NLA_PUT(skb, ODP_KEY_ATTR_8021Q, sizeof(q_key), &q_key);
+		NLA_PUT(skb, OVS_KEY_ATTR_8021Q, sizeof(q_key), &q_key);
 	}
 
 	if (swkey->eth.type == htons(ETH_P_802_2))
 		return 0;
 
-	NLA_PUT_BE16(skb, ODP_KEY_ATTR_ETHERTYPE, swkey->eth.type);
+	NLA_PUT_BE16(skb, OVS_KEY_ATTR_ETHERTYPE, swkey->eth.type);
 
 	if (swkey->eth.type == htons(ETH_P_IP)) {
-		struct odp_key_ipv4 *ipv4_key;
+		struct ovs_key_ipv4 *ipv4_key;
 
-		nla = nla_reserve(skb, ODP_KEY_ATTR_IPV4, sizeof(*ipv4_key));
+		nla = nla_reserve(skb, OVS_KEY_ATTR_IPV4, sizeof(*ipv4_key));
 		if (!nla)
 			goto nla_put_failure;
 		ipv4_key = nla_data(nla);
-		memset(ipv4_key, 0, sizeof(struct odp_key_ipv4));
+		memset(ipv4_key, 0, sizeof(struct ovs_key_ipv4));
 		ipv4_key->ipv4_src = swkey->ipv4.addr.src;
 		ipv4_key->ipv4_dst = swkey->ipv4.addr.dst;
 		ipv4_key->ipv4_proto = swkey->ip.proto;
-		ipv4_key->ipv4_tos = swkey->ip.tos;
+		ipv4_key->ipv4_tos = swkey->ip.tos_frag & ~INET_ECN_MASK;
+		ipv4_key->ipv4_frag = swkey->ip.tos_frag & OVS_FRAG_TYPE_MASK;
 	} else if (swkey->eth.type == htons(ETH_P_IPV6)) {
-		struct odp_key_ipv6 *ipv6_key;
+		struct ovs_key_ipv6 *ipv6_key;
 
-		nla = nla_reserve(skb, ODP_KEY_ATTR_IPV6, sizeof(*ipv6_key));
+		nla = nla_reserve(skb, OVS_KEY_ATTR_IPV6, sizeof(*ipv6_key));
 		if (!nla)
 			goto nla_put_failure;
 		ipv6_key = nla_data(nla);
-		memset(ipv6_key, 0, sizeof(struct odp_key_ipv6));
+		memset(ipv6_key, 0, sizeof(struct ovs_key_ipv6));
 		memcpy(ipv6_key->ipv6_src, &swkey->ipv6.addr.src,
 				sizeof(ipv6_key->ipv6_src));
 		memcpy(ipv6_key->ipv6_dst, &swkey->ipv6.addr.dst,
 				sizeof(ipv6_key->ipv6_dst));
 		ipv6_key->ipv6_proto = swkey->ip.proto;
-		ipv6_key->ipv6_tos = swkey->ip.tos;
+		ipv6_key->ipv6_tos = swkey->ip.tos_frag & ~INET_ECN_MASK;
+		ipv6_key->ipv6_frag = swkey->ip.tos_frag & OVS_FRAG_TYPE_MASK;
 	} else if (swkey->eth.type == htons(ETH_P_ARP)) {
-		struct odp_key_arp *arp_key;
+		struct ovs_key_arp *arp_key;
 
-		nla = nla_reserve(skb, ODP_KEY_ATTR_ARP, sizeof(*arp_key));
+		nla = nla_reserve(skb, OVS_KEY_ATTR_ARP, sizeof(*arp_key));
 		if (!nla)
 			goto nla_put_failure;
 		arp_key = nla_data(nla);
-		memset(arp_key, 0, sizeof(struct odp_key_arp));
+		memset(arp_key, 0, sizeof(struct ovs_key_arp));
 		arp_key->arp_sip = swkey->ipv4.addr.src;
 		arp_key->arp_tip = swkey->ipv4.addr.dst;
 		arp_key->arp_op = htons(swkey->ip.proto);
@@ -1002,13 +1267,14 @@ int flow_to_nlattrs(const struct sw_flow_key *swkey, struct sk_buff *skb)
 		memcpy(arp_key->arp_tha, swkey->ipv4.arp.tha, ETH_ALEN);
 	}
 
-	if (swkey->eth.type == htons(ETH_P_IP) ||
-	    swkey->eth.type == htons(ETH_P_IPV6)) {
+	if ((swkey->eth.type == htons(ETH_P_IP) ||
+	     swkey->eth.type == htons(ETH_P_IPV6)) &&
+	    (swkey->ip.tos_frag & OVS_FRAG_TYPE_MASK) != OVS_FRAG_TYPE_LATER) {
 
 		if (swkey->ip.proto == IPPROTO_TCP) {
-			struct odp_key_tcp *tcp_key;
+			struct ovs_key_tcp *tcp_key;
 
-			nla = nla_reserve(skb, ODP_KEY_ATTR_TCP, sizeof(*tcp_key));
+			nla = nla_reserve(skb, OVS_KEY_ATTR_TCP, sizeof(*tcp_key));
 			if (!nla)
 				goto nla_put_failure;
 			tcp_key = nla_data(nla);
@@ -1020,9 +1286,9 @@ int flow_to_nlattrs(const struct sw_flow_key *swkey, struct sk_buff *skb)
 				tcp_key->tcp_dst = swkey->ipv6.tp.dst;
 			}
 		} else if (swkey->ip.proto == IPPROTO_UDP) {
-			struct odp_key_udp *udp_key;
+			struct ovs_key_udp *udp_key;
 
-			nla = nla_reserve(skb, ODP_KEY_ATTR_UDP, sizeof(*udp_key));
+			nla = nla_reserve(skb, OVS_KEY_ATTR_UDP, sizeof(*udp_key));
 			if (!nla)
 				goto nla_put_failure;
 			udp_key = nla_data(nla);
@@ -1035,9 +1301,9 @@ int flow_to_nlattrs(const struct sw_flow_key *swkey, struct sk_buff *skb)
 			}
 		} else if (swkey->eth.type == htons(ETH_P_IP) &&
 			   swkey->ip.proto == IPPROTO_ICMP) {
-			struct odp_key_icmp *icmp_key;
+			struct ovs_key_icmp *icmp_key;
 
-			nla = nla_reserve(skb, ODP_KEY_ATTR_ICMP, sizeof(*icmp_key));
+			nla = nla_reserve(skb, OVS_KEY_ATTR_ICMP, sizeof(*icmp_key));
 			if (!nla)
 				goto nla_put_failure;
 			icmp_key = nla_data(nla);
@@ -1045,9 +1311,9 @@ int flow_to_nlattrs(const struct sw_flow_key *swkey, struct sk_buff *skb)
 			icmp_key->icmp_code = ntohs(swkey->ipv4.tp.dst);
 		} else if (swkey->eth.type == htons(ETH_P_IPV6) &&
 			   swkey->ip.proto == IPPROTO_ICMPV6) {
-			struct odp_key_icmpv6 *icmpv6_key;
+			struct ovs_key_icmpv6 *icmpv6_key;
 
-			nla = nla_reserve(skb, ODP_KEY_ATTR_ICMPV6,
+			nla = nla_reserve(skb, OVS_KEY_ATTR_ICMPV6,
 						sizeof(*icmpv6_key));
 			if (!nla)
 				goto nla_put_failure;
@@ -1057,9 +1323,9 @@ int flow_to_nlattrs(const struct sw_flow_key *swkey, struct sk_buff *skb)
 
 			if (icmpv6_key->icmpv6_type == NDISC_NEIGHBOUR_SOLICITATION ||
 			    icmpv6_key->icmpv6_type == NDISC_NEIGHBOUR_ADVERTISEMENT) {
-				struct odp_key_nd *nd_key;
+				struct ovs_key_nd *nd_key;
 
-				nla = nla_reserve(skb, ODP_KEY_ATTR_ND, sizeof(*nd_key));
+				nla = nla_reserve(skb, OVS_KEY_ATTR_ND, sizeof(*nd_key));
 				if (!nla)
 					goto nla_put_failure;
 				nd_key = nla_data(nla);
